@@ -9,16 +9,15 @@ from typing import Union
 def freeze_includes() -> List[str]:
     """Return a list of module names used by pytest that should be
     included by cx_freeze."""
-    import py
     import _pytest
 
-    result = list(_iter_all_modules(py))
-    result += list(_iter_all_modules(_pytest))
+    result = list(_iter_all_modules(_pytest))
     return result
 
 
 def _iter_all_modules(
-    package: Union[str, types.ModuleType], prefix: str = "",
+    package: Union[str, types.ModuleType],
+    prefix: str = "",
 ) -> Iterator[str]:
     """Iterate over the names of all modules that can be found in the given
     package, recursively.
